/**
 * This class is factory for all throwable flying object types.
 * In other words fireballs.
 */

class Obstacle extends Piece{

    /**
     * Create the Obstacle.
     *
     * @return {Object} the Obstacle object.
     */
    constructor()
    {
        // Below are all Defaults that may be overriden upon use.
        var width = 108;
        var height = 75;

        var color = "#f1f1f100";

        var x = frame.width;
        var y = rand(55, frame.height/2); //Math.floor(Math.random() * (frame.height/2 ) + 55 );

        // call the piece class constructor.
        super(width,height,color,x,y);

        // type 1 = ball type 2 = bullet
        this.type = 1; // Math.floor(Math.random() * (1 ) + 1);

        //control extra functions such as running and jumping.
        this.bounces =  this.type == 1 ?  rand(1, 4):0 ; 

        this.bounceHeight =  this.type==1 ? -1 * rand(5,12): 0; 

        this.isBouncing = false;

        this.speedX = rand(4,6); 
        this.speedX *= -1;
        this.speedY = 0;
        this.radius = this.height/2;
        this.image = null;
        this.redrawImage=false;
        this.strength = 100;

    }
    /**
     * Returns the current object state.
     * If we use a Spritesheet the point on the sprite to show.
     * @return {coordinate} x and y on the spritesheet to present.
     */
    getState(){

        return this.speedX < 0 ?{x:0, y: 0 }: { x:108 ,y:0 };
    }

    /**
     * Update the objects representatioin on the screen.
     * All obstacles are parametric.
     * 
     * @return {void} nil
     */
    update()
    {
        // uptain the game area context.
        var ctx = myGameArea.context;
        //set the color.
        ctx.fillStyle = this.color;

        // the canvas { x,y} origin is always top left.
        // we set _x, _y to be the center of the parametric { circle, ellipse, etc..}
        this._x = this.x + this.radius;
        this._y = this.y+ this.radius;

        // draw the underlying object.
        ctx.beginPath();
        var state = this.getState();
        ctx.arc(this._x, this._y ,this.radius, 0, 2* Math.PI,false );
        ctx.fill();

        // if we wish to represent with an image superimpose that over the object.
        if (this.redrawImage)  ctx.drawImage(this.image,state.x,state.y, this.width,this.height, this.x, this.y,this.width,this.height);
            
    }
    /**
     * If the bounds of the scene are hit. Run the sequence of events,
     * @return {void} nil
     */
    hitBounds()
    {
        this.hitBottom();
        this.hitTop();

    }
    /**
     * When the bottom of the scene is touched.
     *
     * Initially obstacles were going to be allowed to bounce, but this would make Bowser to difficult to defeat, or Mario have 
     * too easy a time shooting fireballs may be enabled though.
     *
     * @return {void} nil
     */
    hitBottom()
    {
        super.hitBottom();
        if (this.isAtBottom())
        {
            this.bounce();
        }
    }
    /**
     * Sets the x and y position of the object depending on speed and gravity force.
     * Negative gravities are utilized in jumps and bounces.
     *
     * @return {void} nil
     */
    position()
    {
  
        this.gravitySpeed += this.gravity;

   
        if (this.gravitySpeed <= this.bounceHeight )
        {
            this.isBouncing = false;
            this.gravity  = 0.25;

        } 

        this.y += this.speedY + this.gravitySpeed;
        this.x += this.speedX;

        this.hitBounds();

    }
    /**
     * Causes the object to bounce.
     *
     * @return {void} nil
     */
    bounce()
    {

        // cant jump if not at bottom: 
        if (this.bounces <= 0 ) return;
 
        this.bounceHeight /= 2;

        this.gravity = -0.2 ;

        this.isBouncing = true;
        this.bounces--;

    }
}