/**
 * Subclasses the obstacle object to be specific for 
 * Marios spit fire.
 */
class FireSpit extends Obstacle
{

    /**
     * create the spit fire.
     */
    constructor()
    {   
            super();

            
            this.bounces = 0;
            this.gravity = 0.05;
            this.height = 18;
            this.width = 26;
            this.radius = 9;
            this.type = 1;
            this.color = "yellow";
            // this.bounceHeight = -6;
            this.image = new Image();
            this.image.src = "media/fireball.png";
            this.strength = 10;

    }
    /**
     * Update the spitfires scene representation.
     *
     * @return {void} nil
     */
    update()
    {
        if (this.canRender(1) == false) return;
        var ctx = myGameArea.context;

        this._x = this.x + this.radius;
        this._y = this.y+ this.radius;
        ctx.beginPath();
        var state = this.getState();
        ctx.drawImage(this.image,state.x,state.y, 109,75, this.x, this. y,this.width,this.height);
            
        
 
    }
    /**
     * Get the current state of the spite fire from the spritesheet.
     * @return {void} nil
     */
    getState(){
        // determine the point depending on direction.
        return this.speedX < 0 ?{x:0, y: 0 }: { x:108 ,y:0 };
    }
    /**
     * if a bound on the scene is hit do nothing.
     * @return {void} nil
     */
    hitBounds(){
        return;
    }
}

/**
 * This piece is the main protagonist ( Mario)
 * Extends from the Piece class with some overrides.
 */
class Player extends Piece{

    /**
     * Create the piece.
     *
     * @return {Piece} The protagonist Mario..
     */
    constructor(defaults)
    {
        
        var width = 30;
        var height = 55;
        var color = "#f1f1f100";
        var x = 200;
        var y = 0;

        super(width,height,color,x,y);

        // default values when not shown.
        this.defaults = {   
                            walk:1.3, 
                            run:2.3, 
                            lowJump:7.0, 
                            highJump:12.0, 
                            speedX:4.2
                        };

        this.configure(defaults);

        this.type = 0;
        //control extra functions such as running and jumping.

        this.isJumping = false;
        this.spriteX = 0;

        // draw the mario image on the screen.
        this.mario = new Image();
        this.mario.src = "media/mario.png";

        // we are using SpriteSheets..
        // when  forward
        this.fPoints = {
            standing:[
                        {   x:18,  y:35 }
                     ],
            walking: [
                        {   x:79,   y:35    }, {    x:79,   y:126   }, {    x:18,   y:35    }
                     ],
            runningStart: [
                            {   x:590,  y:126   }
                          ],
            running:[
                        {   x:145,  y:35    },{ x:79,   y:126   }, {    x:145,  y:126   }
                    ],
            airborne:[
                        {   x:398,  y:35    }
                     ],
            crouch:[
                        {   x:721,  y:145   }
                   ]
        };
        //draw dead state..
        this.dead = {   x:15,   y:400   };

        //when  backwards..
        this.rPoints = {
            standing:[
                        {   x:15,   y:220   }
                     ],
            walking: [
                        {   x:716,  y:220   }, {    x:716,  y:312   },  {   x:780,  y:220   }
                     ],
            runningStart: [
                                {   x:587,  y:312   }
                        ],
            running:[
                        {   x:654,  y:220   },  {   x:716,  y:312   }, {    x:654,  y:312   }
                    ],
            airborne:[
                        {   x:400,  y:220   }
                     ],
            crouch:[
                        {   x:80,   y:330   }
                   ]
        };  
        this.playerStateIndex = 0;

         // jump music..
         this.jumpSound = new Audio("media/mario-jump.mp3");
         this.jumpSound.volume = .12;
         this.jumpSound.load();
    

        // dies  music..
         this.diesSound = new Audio("media/mario-dies.mp3");
         this.diesSound.volume = .5;
         this.diesSound.load();

         // spit fire sound.
        this.spitSound = new Audio("media/SpitFire.mp3");
        this.spitSound.volume = .5;
        this.spitSound.load();
   
        this.isDead = false;
    }

    /**
     * Obtains the players current state.
     *
     * @return {dictionary} A JSON object that holds key value pairs for the current state.
     */
    getPlayerState()
    {
        if (this.isDead) return this.dead;

        // if pointing forward obtain from the forward dictionary else reverse.
        var states = this.forward ? this.fPoints : this.rPoints;

        if (this.crouch) return states.crouch[0];
      
        if (this.isAtBottom() == false) return states.airborne[0]; 

        // get intermediate states i.e. starting to run...
        if ( this.powerUp && this.playerStateIndex == -1 ) {
            this.playerStateIndex++;
            return states.runningStart[0];
        }

        if (this.speedX !=  0 && this.playerStateIndex >= 0)
        {
            var translation = this.powerUp ? states.running : states.walking;

            // if the array size is exceeded reset to zero.
            this.playerStateIndex = this.playerStateIndex >= translation.length ? 0 : this.playerStateIndex;
            return translation[this.playerStateIndex ];
        } 
        return states.standing[0];

    }
    /**
     * update the players position and image.
     * @return {void} nil
     */
    update()
    {
        var state = this.getPlayerState();
        var ctx = myGameArea.context;
        // create the underlying rectangle.
        ctx.fillStyle = this.color;
        ctx.fillRect (this.x,this.y, this.width,this.height);
        //superimpose his image.
        ctx.drawImage(this.mario,state.x,state.y, 39,55, this.x, this.y,39,55);
        this.spriteX += this.speedX == 0 ? 0: 30;
        this.spriteX = this.spriteX > 120 ? 0: this.spriteX;

        // update every 4 frames, if running then every 2 frames.
        var renderRate = this.powerUp ? 2 : 4;
         if (this.canRender(renderRate)) {
            this.playerStateIndex++;
         }  
    }
    /**
     * run default hitbottom, hit top.
     * then stop him short of the sides.
     * @return {void} nil.
     */
    hitBounds(){
        this.hitBottom();
        this.hitTop();

        // do not allow mario to go past the left boundary.
        this.x = this.x < 0 ? 0 : this.x;
        // do not allow mario to go past the right boundary.
        this.x = (this.x > (frame.width - this.width - 2)) ? frame.width - this.width - 2: this.x;
    }
    /**
     * performs the attacking for Mario.
     * @return {void} nil.
     */
    attack()
    {
        var  x = this.opponent.x;
        var  y = this.opponent.y;

          // can spit only every 250 milliseconds.
         if (this.performAction (0.25) == false )return;

            var fireSpit = new FireSpit();
            fireSpit.speedX = this.forward ? 13: -13;
            fireSpit.x = this.x+ ( this.forward ? 0 : mario.width);
            fireSpit.x = this.x + 30;
            fireSpit.y = this.y + 5 ;

            // play the spit sounds.
            this.play(this.spitSound);
            
            // play the attack on the attack queue.
            myGameArea.marioAttacks.push(fireSpit);
    }
    /**
     * Get Marios position respective to his speed and gravity effects.
     *
     * @return {void} nil
     */
    position()
    {
        // update his vertical speed respective to gravity.
        // * in real life we descend at a force of 9.8 m/s^2
        // that force turns into a velocity which is added to the initial velocity.
        // v = vo + at 
        this.gravitySpeed += this.gravity;
     
        // once his max jump height is reached.
        // switch gravity to positive ( falling).
        if (this.gravitySpeed <= this.jumpHeight)
        {
            this.isJumping = false;
            this.gravity  = 0.35;
        } 
  
        this.x += this.speedX * (this.powerUp ? this.defaults.run : this.defaults.walk); // if we are running our hero can double his speed.
        this.y += this.speedY + this.gravitySpeed;

        // when dead he drops from the screen and no horizontal translation occurs.
        if (this.isDead)
        {
            this.speedX = 0;
        }
        else{
            this.hitBounds();
        }
    }

    /**
     * The functions allows Mario to jump by switching gravity.
     *
     * @return {void} nil
     */
    jump()
    {
        // cant jump if not at bottom: 
        if (this.isAtBottom() == false && this.isDead == false) return;
 
        this.jumpSound.play();
        // control the jump height.
        var level = this.powerUp  ? 2 : 1;
        this.jumpHeight = level == 1 ? this.defaults.lowJump :  this.defaults.highJump;
        this.jumpHeight *= -1;

        // when mario jumps higher he uses more force therefore reaches his height quicker.
        this.gravity = ( 0.35 * level) * -1;

        this.isJumping = true;
    }

    /**
     * Player has lost, sucks.
     *
     * @return {void} nil
     */
    die()
    {
         if (!this.isDead)
         {
            this.diesSound.play()
            this.isDead = true;
            this.gravity = -0.55;
            this.jumpHeight = -10.0;
         }       
    }
    /**
     * Process all keyboard events.
     * @param {String} action whether keydown or keyup.
     * @param {Event} e they keyboard event, holds a code for each individual keystroke.
     */
    processUserInput( action, e )
    {
         if (action == "keydown")
         {
            if(this.crouch){
                this.speedX = 0;
                return;
            } 

                switch (e.keyCode) 
                {
                    //spacebar:
                    case 32:
                       this.powerUp = true;
                       break;
                    // up arrow: 
                    case 38:                  
                        this.jump();
                        break;
                        // left arrow:
                    case 37:
                        this.speedX = -1* this.defaults.speedX;
                        this.forward = false;
                        break;
                        // right arrow.
                    case 39:
                         this.speedX = this.defaults.speedX * 1.0;
                         this.forward = true;
                        break;
                    // down arrow:
                    case 40:
                         this.crouch = true;
                         this.height = 40;
                        break;
                    default:
                        this.attack();
                    }
         }
         else //keyup
         {
                switch (e.keyCode) 
                {
                    //spacebar:
                    case 32:
                       this.powerUp = false;
                       break;
                    // up arrow now big deal on key up..
                    case 38:
                       null;
                       break;
                    // set the crouch flag to false.
                    case 40:
                         this.speedX = 0;
                         this.crouch = false;
                         this.height = 55;
                        break;
                    default:
                         this.speedX = 0;
                    
                    }
         }
    }

}