/**
 * This piece is the main protagonist ( Mario)
 * Extends from the Piece class with some overrides.
 */
class Villain extends Piece{

    /**
     * Create the piece.
     *
     * @return {Piece} The protagonist Mario..
     */
    constructor(defaults)
    {
        var width = 110;
        var height = 90;

        var color = "#f1f1f100";
        var x = frame.width - 2 * height;
        var y = 0;


        super(width,height,color,x,y);

         // default values when not shown.
        this.defaults = {
                            walk: 5.0,
                            run:  7.0,
                            bumrush: 9.0,
                            attackDelay:1.2,
                            fireBallSpeed: 10
                        };

        this.configure(defaults);
        
        this.type = 1;

        //control extra functions such as running and jumping.
        this.isJumping = false;
        this.spriteX = 0;

         // draw the Bowser image on the screen.
        this.bowser = new Image();
        this.bowser.src = "media/bowser.png";
        this.defaultHeight = this.height;
       
        this.playerStateIndex = 0;
        this.renderRate = 2;

        this.stomping = false;

        // dies  music..
        this.death = new Audio("media/bowserDeath.mp3");
        this.death.volume = .4;
        this.death.load();

        // fire music.
        this.fire = new Audio("media/bowserFire.mp3");
        this.fire.volume = .4;
        this.fire.load();

        // is dead.
        this.isDead = false;
        this.lastAttack  = 0;

        this.opponent;
         // we are using SpriteSheets..
        // when  forward
        this.rPoints = {
                    standing:[
                                { x:390, y:250, h:85 }
                        ], 
                    walking: [
                                { x:390, y:250, h:85 },   { x:263, y:250, h:85 },     { x:138,  y:250, h:85 }
                            ],
                    fire:[
                                { x:130,  y:350, h:100 },     { x:5,  y:365,   h:80 }
                        ], 
                    hurt:[
                                {  x:268,   y:350  }
                        ],
                    airborne:[
                                {  x:390,  y:340 , h:100}
                        ],  
                    crouch:[    
                                {   x:16,   y:250, h:80 }
                    ]
                };

        //when  backwards..
        this.fPoints = {
                    standing:[
                                {x:10,y:25,h:85}
                            ],  
                    walking: [
                                {x:10,y:25,h:85},   {x:130,y:25,h:85},  {x:260,y:25, h:85}
                        ],
                    fire:[
                                {x:270,y:125, h:100},   {x:390,y:140,h:80}
                        ],  
                    hurt:[
                                {x:130,y:120}
                         ],
                    airborne:[  
                                {x:7,y:120, h:100}
                    ],  
                    crouch:[
                        { x:380, y:28 , h:80 }
                    ]
                };

    }

    /**
     * Obtains the players current state.
     *
     * @return {dictionary} A JSON object that holds key value pairs for the current state.
     */
    getPlayerState()
    {
        // if pointing forward obtain from the forward dictionary else reverse.
        var states = this.forward ? this.fPoints : this.rPoints;

        if (this.isDead)
        { 
            return states.hurt[0];
        } 
        if (this.crouch) return states.crouch[0];

        if(this.isAirBorne() == true )   return states.airborne[0]; 

        // with bowser when throwing fire 
        // he first opens up then explodes..
        // this must occur on two different frames.
        if(this.throwFire)
        {      
            if(this.playerStateIndex > 1)
            {
             this.throwFire  = false;
             this.renderRate = 2;
             this.playerStateIndex = 0;
             return states.standing[this.playerStateIndex];
            }
            return states.fire[this.playerStateIndex];
        }
        // when moving forward or backwards.
        if (this.speedX !=  0 && this.playerStateIndex >= 0)
        {
            var translation = this.powerUp ? states.running : states.walking;

            this.playerStateIndex = this.playerStateIndex >= translation.length ? 0 : this.playerStateIndex;
            
            return translation[this.playerStateIndex ];
        } 
        return states.standing[0];

    }
    /**
     * update the players position and image.
     * @return {void} nil
     */
    update(){
        var state = this.getPlayerState();
        var ctx = myGameArea.context;

        // different states require different height.
        this.height = state.h || this.height;
        this._x = this.x + this.width/2;
        this._y = this.y +this.height/2 ;

     
        ctx.fillStyle = this.color;

        this.radius = Math.floor(this.width, this.height)/2;

        // ctx.fillRect (this.x,this.y, this.width,this.height);
        ctx.beginPath();
        ctx.arc(this._x, this._y, this.radius, 0, 2* Math.PI,false );
        ctx.fill();

        ctx.drawImage(this.bowser,state.x,state.y, this.width,this.height, this.x, this.y,this.width,this.height);
        ctx.fillStyle = "#fff";
        ctx.font = "20px Arial";
        ctx.fillText("Bowser: ", frame.width - 400, 50);

        // create lifebar border..
        ctx.fillStyle = "#ffffff00";
        ctx.strokeStyle = "fff";
        ctx.strokeRect(frame.width - 300, 32, 200 , 20);

        // create lifebar border..
        ctx.fillStyle = "#fff";
        ctx.fillRect(frame.width - 300, 32, 200 * (this.life/100) , 20);



        // perform all second frame actions..
        this.frameActions();

        if (this.collided(this.opponent))
        {
            myGameArea.frameNo = 0;
            backgroundMusic.pause();
            mario.die();
        }


         if (this.canRender(this.renderRate) == false)  return;

        this.playerStateIndex++;
        // perform all second frame actions..
        this.frameActions();

    }

    /**
     * performs the attacking for Bowser.
     * It utilizes random values to process an attack.
     * Horrid when he throws fire and then jumps, very difficult to evade but thats someone elses problem.
     * @return {void} nil.
     */
    attack()
    {
        var x = this.opponent.x;
        if (this.performAction (this.defaults.attackDelay) == false )return;

    
        if (this.isAirBorne() == true  ) return;

         this.forward = this.x > (x + 20) ? true : false;

        var directionalMultiplier = this.forward ? -1.0 : 1.0;

         //get attack.
        var attack = rand(1,5);

        // prevent multiple attacks.
        this.lastAttack = attack == this.lastAttack ? attack + 1: attack;

        this.playerStateIndex = 0;
        switch (this.lastAttack){
            case 1: //walk attack...
                this.speedX = this.defaults.walk * directionalMultiplier;
                break;
            case 2:
                //bum rush.
                this.speedX = this.defaults.bumrush * directionalMultiplier;
                break;
            case 3:
            // fire ball.
                this.fireAttack();
                break;
            // stomp.
            case 4:
                this.stomping = true;
                this.stomp();
                break;
            default:
               this.speedX = this.defaults.run * directionalMultiplier;
        }

    }
    /**
     * FireAttack , Bowser throws fireballs
     * this is a 2 state function. 
     *
     * @return {void} nil
     */
    fireAttack()
    {
        this.speedX = 0;
        this.throwFire = true;
        //set the render rate to 25, slow down the rendering so the player can see the motions.
        this.renderRate = 25;

        var directionalMultiplier = this.forward ? -1:1;

        // if on the second part of the attack emit the fireball.
        if (this.playerStateIndex > 0 )
        {
            this.throwFire = false;
            var fire = new Obstacle();

            // create the fireball.
            fire.speedX = this.defaults.fireBallSpeed * directionalMultiplier;
            fire.image = new Image();
            fire.image.src = "media/fireball.png";
            fire.bounces = 0;
            fire.gravity = 0;
            fire.x = this.x+ ( this.forward ? 0 : this.width);
            fire.y = this.y - 25 ;
            fire.redrawImage = true;
            myGameArea.bowserAttacks.push(fire);
        }
        else{
            this.fire.play();
        }
    }

    /**
     * Bowser will jump and attempt to stomp on Mario.
     * It calculates the horizontal distance to derive a jump height.
     * that will allow Bowser to overshoot Mario, once his position is over Mario he drops on top of him.
     *
     * @return {void} nil
     */
    stomp()
    {
        // get opponent coordinates.
        var x = this.opponent.x;
        var y = this.opponent.y;

        // get the horizontal distance separation
        var xDelta = Math.abs(this.x - x);
        // get the slope for the jump.
        var slopeCoefficient = (45/360);
        // normalize the slope...
        var slope = slopeCoefficient/ ( 4 * Math.PI);
        //gravity speed.
        var gspeed = 0;
        // how many increments to take...
        var steps = 0;

        this.gravity = -0.35;
        this.jumpHeight =  -1 * xDelta * slope;
        
        // obtain how many increments to reach the jump height in gravity..
        while (gspeed > this.jumpHeight)
        {
            gspeed += this.gravity;
            steps++;
        }


        this.isJumping = true;

        this.speedX = xDelta/(steps * 3 ) ;
        this.speedX *= this.forward ? -1 : 1;

    }

    /**
     * For actions that require multiple frames.
     *
     * @return {void} nil
     */
    frameActions()
    {
        var x = this.opponent.x;
        var y = this.opponent.y;

        if (this.throwFire)
        {
            if (this.playerStateIndex > 0)
            {
                this.fireAttack(x,y);
            }     
        }
        // it would be cool to have Bowser throw Acid bubbles
        // or grenades but that would be too difficult to pass.
    }
    /**
     * run default hitbottom, hit top.
     * then stop him short of the sides.
     * @return {void} nil.
     */
    hitBounds(){
        this.hitBottom();
        this.hitTop();

        // do not allow mario to go past the left boundary.
        this.x = this.x < 0 ? 0 : this.x;
        // do not allow mario to go past the right boundary.
        this.x = (this.x > (frame.width - this.width - 2)) ? frame.width - this.width - 2: this.x;
    }

    /**
     * Get Bowsers position respective to his speed and gravity effects.
     *
     * @return {void} nil
     */
    position()
    {
        // update his vertical speed respective to gravity.
        // * in real life we descend at a force of 9.8 m/s^2
        // that force turns into a velocity which is added to the initial velocity.
        // v = vo +
        this.gravitySpeed += this.gravity;
     
        if (this.gravitySpeed <= this.jumpHeight)
        {
            this.isJumping = false;
            this.gravity  = 0.35;
        } 
  
        this.x += this.speedX * (this.powerUp ? 2.3 : 1); // if we are running our hero can double his speed.
        this.y += this.speedY + this.gravitySpeed;

        // when stomping once Bowser closes on Mario slow down so he lands on him.
        // usually when we jump on something we can slow ourselves by adjusting our bodies.
        // Bowser will be able to do the same.
        if(this.stomping && ( this.x > this.opponent.x && this.x < this.opponent.x + this.opponent.width )) {
            this.speedX = this.forward ? -5.0 : 5.0;
            this.stomping = false;
        }
        if (this.isDead)
        {
            this.speedX = 0;
        }
        else{
            this.hitBounds();
        }
        

    }
    /**
     * The functions allows Bowser to jump by switching gravity.
     *
     * @return {void} nil
     */
    jump()
    {
        // cant jump if not at bottom: 
        if (this.isAtBottom() == false && this.isDead == false) return;
 
        this.jumpSound.play();
        var level = this.powerUp  ? 2 : 1;

        // set the jumpheight.
        this.jumpHeight = level == 1 ? -6.0 : -12.0;

        this.gravity = ( 0.35 * level) * -1;

        this.isJumping = true;

    }
    /** 
     * Kills Bowser..
     *
     * @return {void} nil
     */
    die()
    {
         if (!this.isDead)
         {
            this.death.play()
            this.isDead = true;
            this.gravity = -0.55;
            this.jumpHeight = -10.0;
         }       
    }

}